# -*- coding: utf-8 -*-
'''
Contains the HTTP writer used to communicate with the Asset Performance Management service
'''

# Import Python libs
import asyncio
import os
import os.path
from typing import Any, List, Tuple, Type, TypeVar

# Import local libs
# pylint: disable=import-error,3rd-party-module-not-gated
try:
    from systemlink.messagebus.paths import get_http_master_file
    from systemlink.clientconfig import get_configuration_by_id, HTTP_MASTER_CONFIGURATION_ID
    from systemlink.clients.niapm import (
        ApiClient, ApiException, AssetsApi,
        AssetCreateModel, AssetUpdateModel,
        CreateAssetsRequest, UpdateAssetsRequest
    )
    from systemlink.clients.niapm.rest import RESTResponse
    HAS_SYSTEMLINK_SDK = True
except ImportError:
    HAS_SYSTEMLINK_SDK = False


T = TypeVar('T', bound='AssetPerformanceManagmentHttpWriter')  # pylint: disable=invalid-name


class AssetPerformanceManagmentHttpWriterException(Exception):
    '''
    Exception raised by the AssetPerformanceManagmentHttpWriter class.
    '''

    def __init__(
            self,
            message: str,
            inner_exception: Exception = None,
            is_warning: bool = None) -> None:
        '''
        Initialize the exception.

        :param message: The message of the exception.
        :param inner_exception: The inner exception of the exception.
        :param is_warning: Whether the exception is a warning.
        '''
        if inner_exception:
            msg = '{0}: {1}'.format(message, inner_exception)
        else:
            msg = message

        super().__init__(msg)

        self.is_warning = is_warning is True


class AssetPerformanceManagmentHttpWriter:  # pylint: disable=too-few-public-methods
    '''
    Abstraction over HTTP connection which facilitates communication
    between the minion and the AssetPerformanceManagement service.
    '''
    __instance = None

    @staticmethod
    def has_systemlink_sdk():
        '''
        Check if the module sucessfully loaded NI Skyline Message Bus
        '''
        return HAS_SYSTEMLINK_SDK

    def __new__(cls: Type[T]) -> T:
        '''
        Handle instantiation so that only a single instance of this class is created.
        '''
        if not AssetPerformanceManagmentHttpWriter.has_systemlink_sdk():
            raise AssetPerformanceManagmentHttpWriterException(
                'Import of NI Skyline Message Bus failed.'
            )

        if AssetPerformanceManagmentHttpWriter.__instance is None:
            AssetPerformanceManagmentHttpWriter.__instance = object.__new__(
                cls
            )
            # pylint: disable=protected-access
            AssetPerformanceManagmentHttpWriter.__instance._initialized = False
            AssetPerformanceManagmentHttpWriter.__instance._event_loop = None
            AssetPerformanceManagmentHttpWriter.__instance._api_client = None
            # pylint: enable=protected-access
        return AssetPerformanceManagmentHttpWriter.__instance

    def __init__(self) -> None:
        '''
        Initialize the message service if not already initialized
        '''
        if not self._initialized:
            try:
                self._setup_communication()
            except ApiException as exc:
                self._cleanup_communication()
                # All HTTP exceptions
                raise AssetPerformanceManagmentHttpWriterException(
                    'An HTTP error has occurred',
                    exc
                )

    def _setup_communication(self) -> None:
        '''
        Set up the communication.
        '''
        file_path = get_http_master_file()
        if not os.path.isfile(file_path):
            # The Skyline Master file is not available.
            # Can't set up the Message Service without it.
            return

        self._setup_http_connection()

        self._initialized = True

    def _setup_http_connection(self) -> None:
        '''
        Set up the HTTP communication and initialize the Assets API object.
        '''
        # These class members are defined in func:`AssetPerformanceManagmentHttpWriter.__new__`
        # instead of func:`AssetPerformanceManagmentHttpWriter.__init__`.
        self._event_loop = asyncio.get_event_loop()  # pylint: disable=attribute-defined-outside-init

        assets_route_name = '/niapm'
        configuration = get_configuration_by_id(
            HTTP_MASTER_CONFIGURATION_ID,
            assets_route_name,
            False)
        # These class members are defined in func:`AssetPerformanceManagmentHttpWriter.__new__`
        # instead of func:`AssetPerformanceManagmentHttpWriter.__init__`.
        self._api_client = ApiClient(configuration=configuration)  # pylint: disable=attribute-defined-outside-init

    def _cleanup_communication(self) -> None:
        '''
        Clean up the HTTP communcation.
        '''
        self._event_loop.run_until_complete(self._api_client.close())
        self._api_client = None
        self._event_loop = None
        self._initialized = False

    @classmethod
    def cleanup(cls: Type[T]) -> None:
        '''
        Clean up the writer by closing the connection
        '''
        # This makes sure that if the salt-minion closes before the writer has been initialized
        # we don't create it just to clean it up.
        if cls.__instance is not None:
            cls.__instance._cleanup_communication()  # pylint: disable=protected-access
            cls.__instance = None

    def update_assets(
            self,
            asset_update_models: List[AssetUpdateModel]) -> List[AssetUpdateModel]:
        '''
        Publish an /update-assets request over HTTP.

        :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
        :return: A list of AssetUpdateModels objects that were not found on the server.
        '''
        try:
            not_found_error_code = -254772

            update_assets_request = UpdateAssetsRequest(
                asset_update_models
            )
            assets_api = AssetsApi(api_client=self._api_client)
            task = asyncio.ensure_future(assets_api.update_assets(
                update_assets_request,
                _preload_content=False))
            update_assets_response = self._run_request(
                task,
                deserialize=True,
                success_response_type='UpdateAssetsPartialSuccessResponse')

            failed_assets = update_assets_response.failed
            error = update_assets_response.error
            if error is None:
                return []
            inner_errors = error.inner_errors
            if inner_errors is None:
                return []
            not_found_assets = []
            for i, inner_error in enumerate(inner_errors):
                if inner_error.code == not_found_error_code:
                    not_found_assets.append(failed_assets[i])
            return not_found_assets
        except ApiException as exc:
            if exc.status == 401:
                # The salt-master may have changed credentials. Notify this accordingly.
                raise AssetPerformanceManagmentHttpWriterException(
                    'HTTP Authentication error. Credentials may have changed',
                    exc
                )
            # All other HTTP exceptions. Notify this accordingly.
            raise AssetPerformanceManagmentHttpWriterException(
                'An HTTP error has occurred',
                exc
            )

    def _run_request(
            self,
            request_task: asyncio.Task,
            allowed_statuses: Tuple[int] = (200,),
            deserialize: bool = False,
            success_response_type: str = None) -> Any:
        '''
        Executes the request and retrieves the deserialized response from the result of the task.

        :param request_task: The request to be executed as a task.
        :param deserialize: Whether the response should be deserialized.
        :param success_response_type: The name of the expected response type for success.
        :return: The response as an object of the specified type or the raw response.
        '''
        self._event_loop.run_until_complete(request_task)
        raw_response = request_task.result()

        data = raw_response.content.read_nowait()
        if raw_response.status not in allowed_statuses:
            rest_response = RESTResponse(raw_response, data)
            raise ApiException(http_resp=rest_response)

        if not deserialize:
            return raw_response

        raw_response.data = data.decode("utf-8")
        deserialized_response = self._api_client.deserialize(
            raw_response, success_response_type
        )
        return deserialized_response

    def create_assets(self, asset_create_models: List[AssetCreateModel]) -> None:
        '''
        Publish an /create-assets request over HTTP.

        :param asset_create_models: A list of AssetCreateModel objects representing the assets.
        '''
        try:
            create_assets_request = CreateAssetsRequest(
                asset_create_models
            )
            assets_api = AssetsApi(api_client=self._api_client)
            task = asyncio.ensure_future(assets_api.create_assets(
                create_assets_request,
                _preload_content=False))
            self._run_request(task)
        except ApiException as exc:
            if exc.status == 401:
                # The salt-master may have changed credentials. Notify this accordingly.
                raise AssetPerformanceManagmentHttpWriterException(
                    'HTTP Authentication error. Credentials may have changed',
                    exc
                )
            # All other HTTP exceptions. Notify this accordingly.
            raise AssetPerformanceManagmentHttpWriterException(
                'An HTTP error has occurred',
                exc
            )
