# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Assets Module
'''
# Import Python libs
import atexit
from datetime import datetime
import fnmatch
try:
    from dateutil.parser import parse
    from dateutil.relativedelta import relativedelta
    HAS_DATEUTIL = True
except ImportError:
    HAS_DATEUTIL = False
import json
import logging
import os
import os.path
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

try:
    import winreg
except ImportError:
    pass  # We are on Linux, this is not an issue

# Import salt libs
# pylint: disable=import-error,3rd-party-module-not-gated
import salt.utils.files
import salt.utils.platform
# pylint: enable=import-error,3rd-party-module-not-gated

# Import local libs
# This file may be loaded out of __pycache__, so the
# directory of its .py may not be in the search path.
IMPORT_PATH = os.path.dirname(__file__)
if IMPORT_PATH.endswith('__pycache__'):
    IMPORT_PATH = os.path.dirname(IMPORT_PATH)
sys.path.append(IMPORT_PATH)
try:
    from _apm_amqp_writer import AssetPerformanceManagmentHttpWriter
    from _apm_amqp_writer import AssetPerformanceManagmentHttpWriterException
    import _apm_constants as apm_constants
finally:
    # Remove the extra search path that we added to sys.path
    sys.path.remove(IMPORT_PATH)

try:
    from systemlink.clients.niapm import (
        AssetLocationWithPresenceModel, AssetType, AssetCreateModel,
        AssetUpdateModel, AssetPresenceModel, AssetPresenceStatus,
        BusType, CalibrationMode, ExternalCalibrationOperatorModel,
        ExternalCalibrationWithChecksumModel, SelfCalibrationModel,
        TemperatureSensorModel
    )
    HAS_SYSTEMLINK_SDK = True
except ImportError:
    HAS_SYSTEMLINK_SDK = False

__virtualname__ = 'nisysmgmt_assets'

# Set up logging
log = logging.getLogger(__name__)

ATEXIT_REGISTERED = False

# Version indicating the version of AssetPerformanceManagementMinionAssetsUpdatedBroadcast
MESSAGE_VERSION = 1

NI_INSTALLERS_REG_PATH = 'SOFTWARE\\National Instruments\\Common\\Installer'
NI_INSTALLERS_REG_KEY_APP_DATA = 'NIPUBAPPDATADIR'

USER_DEFINED_ENABLE_FILE_PATH = None
USER_DEFINED_DIRECTORY_PATH = None

BUS_TYPE_VALUES = {
    'BuiltInSystem': BusType.BUILT_IN_SYSTEM,
    'PciPxi': BusType.PCI_PXI,
    'Usb': BusType.USB,
    'Gpib': BusType.GPIB,
    'Vxi': BusType.VXI,
    'Serial': BusType.SERIAL,
    'TcpIp': BusType.TCP_IP,
    'cRio': BusType.CRIO,
    'Scxi': BusType.SCXI,
    'cDaq': BusType.CDAQ,
    'SwitchBlock': BusType.SWITCH_BLOCK,
    'Scc': BusType.SCC,
    'FireWire': BusType.FIRE_WIRE,
    'Accessory': BusType.ACCESSORY,
    'Can': BusType.CAN,
    'SwitchBlockDevice': BusType.SWITCH_BLOCK_DEVICE
}

ASSET_TYPE_VALUES = {
    'Generic': AssetType.GENERIC,
    'DeviceUnderTest': AssetType.DEVICE_UNDER_TEST,
    'Fixture': AssetType.FIXTURE,
    'System': AssetType.SYSTEM,
}

ASSET_PRESENCE_VALUES = {
    'Initializing': AssetPresenceStatus.INITIALIZING,
    'Unknown': AssetPresenceStatus.UNKNOWN,
    'NotPresent': AssetPresenceStatus.NOT_PRESENT,
    'Present': AssetPresenceStatus.PRESENT,
}


def __virtual__() -> Union[str, Tuple[bool, str]]:
    '''
    During module lazy loading, will return the virtual name of this module.

    :return: The virtual name of this module. On error, return a 2-tuple with
        ``False`` for the first item and the error message for the second.
    '''
    has_systemlink_sdk = AssetPerformanceManagmentHttpWriter.has_systemlink_sdk()
    if not has_systemlink_sdk or not HAS_DATEUTIL or not HAS_SYSTEMLINK_SDK:
        return False, 'This module requires NI Skyline Message Bus and the python-dateutil module for Python'

    return __virtualname__


def publish_user_defined_assets() -> None:
    '''
    Function used to retrieve user-specified assets and asset-related information
    from the Data\\Assets\\UserDefined directory on the minion.
    '''
    try:
        _register_atexit_if_not_registered()

        file_paths = _get_all_json_file_paths_recursively()
        feature_enabled = _is_user_defined_assets_feature_enabled()
        if not file_paths and not feature_enabled:
            # The Data\Assets\UserDefined directory does not exist or it is empty -> NoOp
            return apm_constants.PublishUserDefinedAssetsStatusCode.NO_JSON_FILES.value

        asset_bags = _read_all_assets_from_json_files(file_paths)
        assets = _parse_assets_from_bag(asset_bags)
        # Create a file on disk to make sure from this point on we send messages to the
        # server even if no JSON files exist under the UserDefined directory.
        _enable_user_defined_feature_if_not_enabled(feature_enabled)
        _publish_assets(assets)
    except AssetPerformanceManagmentHttpWriterException as exc:
        if exc.is_warning:
            log.warning(exc)
        else:
            log.error(exc)
        AssetPerformanceManagmentHttpWriter.cleanup()
        return apm_constants.PublishUserDefinedAssetsStatusCode.HTTP_ERROR.value
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "nisysmgmt_assets module": %s',
            exc, exc_info=True
        )
        AssetPerformanceManagmentHttpWriter.cleanup()
        return apm_constants.PublishUserDefinedAssetsStatusCode.INTERNAL_ERROR.value

    return apm_constants.PublishUserDefinedAssetsStatusCode.SUCCESS.value


def _is_user_defined_assets_feature_enabled() -> bool:  # pylint: disable=invalid-name
    '''
    Indicate whether the user defined assets feature is enabled by verifying
    if the .enable file exists on the expected app data directory.

    :return: ``True`` if the feature is enabled, else ``False``.
    '''
    enable_file_path = _get_user_defined_enable_file_path()
    return os.path.isfile(enable_file_path)


def _get_user_defined_enable_file_path() -> str:  # pylint: disable=invalid-name
    '''
    Gets the file path of the .enable file under the expected app data directory.

    :return: The file path of the .enable file.
    '''
    global USER_DEFINED_ENABLE_FILE_PATH  # pylint: disable=global-statement

    if not USER_DEFINED_ENABLE_FILE_PATH:
        common_appdata_dir = _get_skyline_appdata_dir()
        USER_DEFINED_ENABLE_FILE_PATH = os.path.join(
            common_appdata_dir,
            apm_constants.DATA_DIRECTORY,
            apm_constants.ASSETS_DIRECTORY,
            apm_constants.USER_DEFINED_ENABLE_FILE
        )

    return USER_DEFINED_ENABLE_FILE_PATH


def _get_skyline_appdata_dir() -> str:
    '''
    Gets the Skyline AppData directory based on the current platform.

    :return: The path of the Skyline AppData directory.
    '''
    if salt.utils.platform.is_windows():
        return _get_windows_skyline_appdata_dir()
    return _get_linux_skyline_appdata_dir()


def _get_windows_skyline_appdata_dir() -> str:  # pylint: disable=invalid-name
    '''
    Gets the Skyline AppData directory on Windows.

    This looks like: 'C:\\ProgramData\\National Instruments\\Skyline'

    :return: The path of the Skyline AppData directory.
    '''
    with winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE,
            NI_INSTALLERS_REG_PATH,
            0,
            winreg.KEY_READ) as hkey:
        (appdata_dir, _) = winreg.QueryValueEx(
            hkey, NI_INSTALLERS_REG_KEY_APP_DATA)
        return os.path.join(
            appdata_dir,
            'Skyline'
        )


def _get_linux_skyline_appdata_dir() -> str:
    '''
    Gets the Skyline AppData directory on Windows.

    This looks like: '/etc/natinst/niskyline'

    :return: The path of the Skyline AppData directory.
    '''
    return '/etc/natinst/niskyline'


def _register_atexit_if_not_registered() -> None:  # pylint: disable=invalid-name
    '''
    Register cleanup functions.
    '''
    global ATEXIT_REGISTERED  # pylint: disable=global-statement

    if not ATEXIT_REGISTERED:
        atexit.register(_cleanup_writer)
        ATEXIT_REGISTERED = True


def _cleanup_writer() -> None:
    '''
    Call the writer cleanup function
    '''
    AssetPerformanceManagmentHttpWriter.cleanup()


def _get_all_json_file_paths_recursively() -> List[str]:  # pylint: disable=invalid-name
    '''
    Looks at the Data\\Assets\\UserDefined directory from the National Instruments Common Data Directory,
    and identifies all JSON files, returning their absoulte paths.

    :return: The paths of all JSON files under the UserDefined directory.
    '''
    file_paths = []

    assets_dir = _get_user_defined_directory_path()
    for root, _, filenames in os.walk(assets_dir):
        for filename in fnmatch.filter(filenames, '*.json'):
            file_paths.append(os.path.join(root, filename))

    return file_paths


def _get_user_defined_directory_path() -> str:  # pylint: disable=invalid-name
    '''
    Return the directory path of the user defined directory under the expected app data directory.

    :return: The path of the UserDefined directory.
    '''
    global USER_DEFINED_DIRECTORY_PATH  # pylint: disable=global-statement

    if not USER_DEFINED_DIRECTORY_PATH:
        common_appdata_dir = _get_skyline_appdata_dir()
        USER_DEFINED_DIRECTORY_PATH = os.path.join(
            common_appdata_dir,
            apm_constants.DATA_DIRECTORY,
            apm_constants.ASSETS_DIRECTORY,
            apm_constants.USER_DEFINED_DIRECTORY
        )

    return USER_DEFINED_DIRECTORY_PATH


def _read_all_assets_from_json_files(file_paths: List[str]) -> List[Dict[str, Any]]:  # pylint: disable=invalid-name
    '''
    Given a list of file paths to JSON files, extracts the contents of the assets list
    and merges them into a single list.

    :param file_paths: The paths of the JSON files from which to read the asset properties.
    :return: A list of dictionaries, each containing the properties for an asset.
    '''
    assets = []

    for file_path in file_paths:
        try:
            file_content = None

            with salt.utils.files.fopen(file_path) as json_file:
                file_content = json.load(json_file)

            if file_content:
                file_assets = file_content['assets']
                if not _islist(file_assets):
                    continue
                assets.extend(file_assets)
        except (IOError, ValueError):
            log.warning(
                'JSON file %s could not be accessed or it was not a valid JSON. File will be skipped.',
                file_path
            )
        except KeyError:
            log.warning(
                'JSON file %s does not contain an assets list. File will be skipped.',
                file_path
            )

    return assets


def _islist(data: Any) -> bool:
    '''
    Check whether data is a list.

    :param data: The value to be checked.
    :return: ``True`` if data is a list, else ``False``.
    '''
    return _isinstance(data, list, False)


def _isinstance(data: Any, expected_type: type, inheritance_lookup: bool = True) -> bool:
    '''
    This is our own implementation of isinstance(), since in some cases
    we don't want to take inheritance into account when parsing data.

    For example: `isinstance(False, int)` returns True, but we want to make
    sure the user specifed the exact type we expect, which is int. Checking
    `type(False) is int` returns False.

    When you call this as you would call the Python isinstance() function,
    this will behave the same. The inheritance_lookup parameter is set
    by default to True, so it will use the isinstance() function.

    When you call this by specifying the inheritance_lookup parameter
    to False, this will check the data's exact type against the
    expected_type passed in.

    :param data: The value to be checked.
    :param expected_type: The type to verify against which to verify data.
    :param inheritance_lookup: Whether to take inheritance into account.
    :return: ``True`` if data is of the specified type, else ``False``.
    '''
    if inheritance_lookup:
        return isinstance(data, expected_type)

    return type(data) is expected_type  # pylint: disable=unidiomatic-typecheck


def _parse_assets_from_bag(asset_bags: List[Dict[str, Any]]) -> List[AssetUpdateModel]:
    '''
    Given a list of asset properties, builds a list of AssetUpdateModel objcets.

    :param asset_bags: A list of dictionaries, each containing the properties for an asset.
    :return: A list of AssetUpdateModel objects representing the assets.
    '''
    update_asset_models = []

    for asset_bag in asset_bags:
        if not _isdict(asset_bag):
            continue
        properties = asset_bag.copy()
        update_asset_model = _create_asset_with_identification_properties(
            properties)
        if update_asset_model:
            _set_location_properties_on_asset(properties, update_asset_model)
            _set_self_calibration_properties_on_asset(
                properties, update_asset_model)
            _set_external_calibration_properties_on_asset(
                properties, update_asset_model)
            _set_handled_properties_on_asset(properties, update_asset_model)
            update_asset_models.append(update_asset_model)

    return update_asset_models


def _isdict(data: Any) -> bool:
    '''
    Check whether data is a dict.

    :param data: The value to be checked.
    :return: ``True`` if data is a dict, else ``False``.
    '''
    return _isinstance(data, dict, False)


def _create_asset_with_identification_properties(properties: Dict[str, Any]) -> AssetUpdateModel:  # pylint: disable=invalid-name
    '''
    Extracts and validates identification properties from the bag returning an Asset
    with the identification properties set.

    If any of the properties are not present in the bag or are invalid returns None.

    :param properties: A dictionary containing the properties for an asset.
    :return: The AssetUpdateModel object representing the asset.
    '''
    model_number, model_name, has_model = _get_number_and_name_paired_properties(
        properties, apm_constants.MODEL_NUMBER_KEY, apm_constants.MODEL_NAME_KEY
    )

    vendor_number, vendor_name, has_vendor = _get_number_and_name_paired_properties(
        properties, apm_constants.VENDOR_NUMBER_KEY, apm_constants.VENDOR_NAME_KEY
    )

    serial_number = properties.pop(
        _camelcase(apm_constants.SERIAL_NUMBER_KEY), None)
    if not _isstr(serial_number):
        serial_number = None

    bus_type_property = properties.pop(
        _camelcase(apm_constants.BUS_TYPE_KEY), None)
    bus_type = BUS_TYPE_VALUES.get(bus_type_property, BusType.BUILT_IN_SYSTEM)

    if not _is_asset_valid(has_model, has_vendor):
        return None

    return AssetUpdateModel(
        model_name,
        model_number,
        serial_number,
        vendor_name,
        vendor_number,
        bus_type
    )


def _get_number_and_name_paired_properties(  # pylint: disable=invalid-name
        properties: Dict[str, Any],
        number_key: str,
        name_key: str) -> Tuple[int, str, bool]:
    '''
    Extracts the values for the given number_key and name_key from the properties and validates them,
    returning both values and a bool indicating success. The operation is considered successful
    if at least one of the propeties exists and has a valid value.

    :param properties: A dictionary containing the properties for an asset.
    :param properties: The key of the number property.
    :param properties: The key of the name property.
    :return: A tuple containing the number and name property values and a flag indicating success.
    '''
    number = properties.pop(_camelcase(number_key), None)
    if not _isint(number):
        number = None

    name = properties.pop(_camelcase(name_key), None)
    if not _isstr(name):
        name = None

    has_paired_properties = number is not None or name is not None

    return number, name, has_paired_properties


def _camelcase(value: str) -> str:
    '''
    Given a PascalCase value, it converts it to camelCase.

    :param value: The PascaCase input value.
    :return: The camleCase converted value.
    '''
    return value[0].lower() + value[1:]


def _isint(data: Any) -> bool:
    '''
    Check whether data is an int.

    :param data: The value to be checked.
    :return: ``True`` if data is an int, else ``False``.
    '''
    return _isinstance(data, int, False)


def _isstr(data: Any) -> bool:
    '''
    Check whether data is a str.

    :param data: The value to be checked.
    :return: ``True`` if data is a str, else ``False``.
    '''
    return _isinstance(data, str, False)


def _to_str(value: Any) -> str:
    '''
    Converts any value or objects to its string representation.

    :param data: The value to be converted.
    :return: The converted value.
    '''
    # For now we do a simple str() conversion, but in the future we might want to improve
    # it by using salt.utils.stringutils.to_str().
    return str(value)


def _is_asset_valid(has_model: bool, has_vendor: bool) -> bool:
    '''
    Checks that all identification properties exist.

    :param has_model: Whether there are valid model properties.
    :param has_vendor: Whether there are valid vendor properties.
    :return: ``True`` if the asset is valid, else ``False``.
    '''
    return has_model and has_vendor


def _set_location_properties_on_asset(  # pylint: disable=invalid-name
        properties: Dict[str, Any],
        asset: AssetUpdateModel) -> None:
    '''
    Extracts the location property from the bag and inserts the location properties
    in the properties list of the asset.

    Properties include:
    - resource uri
    - slot number
    - parent

    :param properties: A dictionary containing the properties for an asset.
    :param asset: An AssetUpdateModel object representing the asset.
    '''
    minion_id = __grains__['id']

    location = properties.pop(_camelcase(apm_constants.LOCATION_KEY), None)
    if not _isdict(location):
        asset.location = AssetLocationWithPresenceModel(
            minion_id,
            state=AssetPresenceModel(AssetPresenceStatus.PRESENT)
        )
        return

    resource_uri = location.pop(_camelcase(apm_constants.RESOURCE_URI_KEY), None)

    slot_number = location.pop(_camelcase(apm_constants.SLOT_NUMBER_KEY), None)
    if not _isint(slot_number):
        slot_number = None

    parent = location.pop(_camelcase(apm_constants.PARENT_KEY), None)

    # This ensures backwards compatibility. All existing assets from JSON will be considered present
    asset_presence = AssetPresenceStatus.PRESENT
    state_property = location.pop(_camelcase(apm_constants.STATE_KEY), None)
    if _isdict(state_property):
        asset_presence_property = state_property.pop(
            _camelcase(apm_constants.ASSET_PRESENCE_KEY), None)
        asset_presence = ASSET_PRESENCE_VALUES.get(
            asset_presence_property,
            AssetPresenceStatus.UNKNOWN)
    state = AssetPresenceModel(asset_presence)

    asset.location = AssetLocationWithPresenceModel(
        minion_id,
        parent,
        resource_uri,
        slot_number,
        state
    )


def _set_self_calibration_properties_on_asset(  # pylint: disable=invalid-name
        properties: Dict[str, Any],
        asset: AssetUpdateModel) -> None:
    '''
    Extracts the self calibration property from the bag and inserts the calibration properties
    in the properties list of the asset.

    Properties include:
    - calibration date
    - limited

    If the calibration date is invalid or if it does not exist this is a NoOp.

    :param properties: A dictionary containing the properties for an asset.
    :param asset: An AssetUpdateModel object representing the asset.
    '''
    self_calibration = properties.pop(
        _camelcase(apm_constants.SELF_CALIBRATION_KEY), None)
    if not _isdict(self_calibration):
        return

    date_str = self_calibration.pop(
        _camelcase(apm_constants.CALIBARTION_DATE_KEY), None)
    date = _parse_date(date_str)
    if not date:
        return

    limited = self_calibration.pop(
        _camelcase(apm_constants.IS_LIMITED_KEY), None)
    if not _isbool(limited):
        limited = None

    asset.self_calibration = SelfCalibrationModel(
        [],  # We don't support temperature sensors yet in the JSON schema
        limited,
        date
    )


def _parse_date(serialized_date: str) -> datetime:
    '''
    Parses a date given as a string into a datetime.

    :param serialized_date: The date represented as a string.
    :return: The parsed date in datetime format.
    '''
    try:
        return parse(serialized_date)
    except Exception:  # pylint: disable=broad-except
        pass

    return None


def _isbool(data) -> bool:
    '''
    Check whether data is a bool.

    :param data: The value to be checked.
    :return: ``True`` if data is a bool, else ``False``.
    '''
    return _isinstance(data, bool, False)


def _set_external_calibration_properties_on_asset(  # pylint: disable=invalid-name, too-many-locals
        properties: Dict[str, Any],
        asset: AssetUpdateModel) -> None:
    '''
    Extracts the external calibration property from the bag and inserts the calibration properties
    in the properties list of the asset.

    Properties include:
    - calibration date
    - next recommended date
    - recommended interval
    - limited
    - comments
    - checksum
    - operator

    If either one of the calibration date, recommended interval or next recommended date
    are invalid or if they do not exist in the bag this is a NoOp.

    :param properties: A dictionary containing the properties for an asset.
    :param asset: An AssetUpdateModel object representing the asset.
    '''
    external_calibration = properties.pop(
        _camelcase(apm_constants.EXTERNAL_CALIBRATION_KEY), None)
    if not _isdict(external_calibration):
        return

    date_str = external_calibration.pop(
        _camelcase(apm_constants.CALIBARTION_DATE_KEY), None)
    date = _parse_date(date_str)
    interval = external_calibration.pop(
        _camelcase(apm_constants.RECOMMENDED_INTERVAL_KEY), None)
    if not date or not _is_interval_valid(interval):
        return

    next_date = date + relativedelta(months=+interval)

    calibration_type = external_calibration.pop(
        _camelcase(apm_constants.CALIBRATION_TYPE), None)
    calibration_type = _get_calibration_type_or_default(calibration_type, CalibrationMode.MANUAL)
    comments = external_calibration.pop(_camelcase(apm_constants.COMMENTS_KEY), None)
    checksum = external_calibration.pop(_camelcase(apm_constants.CHECKSUM_KEY), None)

    temp_sensors = external_calibration.pop(
        _camelcase(apm_constants.TEMPERATURE_SENSORS), None)
    temp_sensors = _parse_temperature_sensors(temp_sensors)

    limited = external_calibration.pop(
        _camelcase(apm_constants.IS_LIMITED_KEY), None)
    if not _isbool(limited):
        limited = None

    operator = None
    operator_property = external_calibration.pop(_camelcase(apm_constants.OPERATOR_KEY), None)
    if _isdict(operator_property):
        display_name = operator_property.pop(_camelcase(apm_constants.DISPLAY_NAME_KEY), None)
        user_id = operator_property.pop(_camelcase(apm_constants.USER_ID_KEY), None)
        operator = ExternalCalibrationOperatorModel(
            display_name,
            user_id
        )

    asset.external_calibration = ExternalCalibrationWithChecksumModel(
        temp_sensors,
        limited,
        date,
        interval,
        next_date,
        comments,
        calibration_type,
        operator,
        checksum
    )


def _is_interval_valid(data: Any) -> bool:
    '''
    Check whether data is a valid calibration interval.

    :param data: The value to be checked.
    :return: ``True`` if data is a valid interval, else ``False``.
    '''
    return _isint(data) and int(data) > 0


def _get_calibration_type_or_default(calibration_type: Optional[str], default: str) -> str:  # pylint: disable=invalid-name
    '''
    Determine whether the calibration type passed as a parameter is valid.
    If valid, return the value, otherwise return the default value.
    In order for a calibration type to be valid it must be equal either to
    MANUAL or AUTOMATIC (case sensitive).

    :param calibration_type: Nullable value for the calibration type.
    :param default: Default value for the calibration type.
    :return: The calibration type passed as a parameter if the value is valid
    else the default value.
    '''
    if calibration_type in [CalibrationMode.MANUAL, CalibrationMode.AUTOMATIC]:
        return calibration_type
    return default


def _parse_temperature_sensors(sensors: Optional[List[Dict[str, Any]]]) -> List[TemperatureSensorModel]:
    '''
    Parses a list of dictionaries to a list of temperature sensor models.

    :param sensors: A list of dictionaries containing temperature sensors data.
    :return: Empty list if the sensors list is NIL, the parsed list otherwise.
    '''
    if sensors is None:
        return []
    return [sm for sm in [_parse_temperature_sensor(s) for s in sensors] if sm is not None]


def _parse_temperature_sensor(sensor: Dict[str, Any]) -> Optional[TemperatureSensorModel]:
    '''
    Parses a dictionary to a TemperatureSensorModel.
    A temperature sensor model is only valid if the reading is a non null number.

    :param sensor: A dictionary containing property of a temperature sensor model.
    :return: The parsed model or none if the data is invalid.
    '''
    reading = sensor.pop(_camelcase(apm_constants.SENSOR_READING), None)
    if not isinstance(reading, int) and not isinstance(reading, float):
        return None
    name = sensor.pop(_camelcase(apm_constants.SENSOR_NAME), None)
    return TemperatureSensorModel(name, reading)


def _set_handled_properties_on_asset(  # pylint: disable=invalid-name
        properties: Dict[str, Any],
        asset: AssetUpdateModel) -> None:
    '''
    Extracts all handled properties from the bag (except for identification, asset class and self/external calibration)
    and inserts them in the properties list of the asset.

    :param properties: A dictionary containing the properties for an asset.
    :param asset: An AssetUpdateModel object representing the asset.
    '''
    asset_type_property = properties.pop(_camelcase(apm_constants.ASSET_TYPE_KEY), None)
    asset_type = ASSET_TYPE_VALUES.get(asset_type_property, None)

    asset_name = properties.pop(_camelcase(apm_constants.ASSET_NAME_KEY), None)
    firmware_version = properties.pop(_camelcase(apm_constants.FIRMWARE_VERSION_KEY), None)
    hardware_version = properties.pop(_camelcase(apm_constants.HARDWRE_VERSION_KEY), None)
    visa_resource_name = properties.pop(_camelcase(apm_constants.VISA_RESOURCE_NAME_KEY), None)

    supports_self_calibration = properties.pop(
        _camelcase(apm_constants.SUPPORTS_SELF_CALIBRATION_KEY), None)
    if not _isbool(supports_self_calibration):
        supports_self_calibration = None

    supports_external_calibration = properties.pop(
        _camelcase(apm_constants.SUPPORTS_EXTERNAL_CALIBRATION_KEY), None)
    if not _isbool(supports_external_calibration):
        supports_external_calibration = None

    asset.name = asset_name
    asset.asset_type = asset_type
    asset.firmware_version = firmware_version
    asset.hardware_version = hardware_version
    asset.visa_resource_name = visa_resource_name
    asset.supports_self_calibration = supports_self_calibration
    asset.supports_external_calibration = supports_external_calibration
    asset.temperature_sensors = []  # We don't support temperature sensors yet in the JSON schema


def _pascalcase(value: str) -> str:
    '''
    Given a camelCase value, it converts it to PascalCase.

    :param value: The camelCase input value.
    :return: The PascalCase converted value.
    '''
    return value[0].upper() + value[1:]


def _enable_user_defined_feature_if_not_enabled(feature_enabled: bool) -> None:  # pylint: disable=invalid-name
    '''
    Enable the user defined assets feature by writing the .enable file on the expected app data directory.

    :param feature_enables: Whetere the feature is enabled.
    '''
    if feature_enabled:
        return

    enable_file_path = _get_user_defined_enable_file_path()
    with salt.utils.files.fopen(enable_file_path, 'w'):
        pass


def _publish_assets(asset_update_models: List[AssetUpdateModel]) -> None:
    '''
    Call the writer to publish the assets over HTTP.

    :param asset_update_models: A list of AssetUpdateModel objects representing the assets.
    '''
    if not asset_update_models:
        return

    writer = AssetPerformanceManagmentHttpWriter()
    not_found_assets = writer.update_assets(asset_update_models)
    if not_found_assets:
        asset_create_models = _convert_update_to_create_models(not_found_assets)
        writer.create_assets(asset_create_models)


def _convert_update_to_create_models(  # pylint: disable=invalid-name
        asset_update_models: List[AssetUpdateModel]) -> List[AssetCreateModel]:
    '''
    Convertes a list of AssetUpdateModel objects into AssetCreateModel objects.

    :param asset_update_models: The input list of AssetUpdateModel objects.
    :return: The converted list of AssetCreateModel objects.
    '''
    asset_create_models = []
    for asset_update_model in asset_update_models:
        asset_create_model = AssetCreateModel(
            asset_update_model.model_name,
            asset_update_model.model_number,
            asset_update_model.serial_number,
            asset_update_model.vendor_name,
            asset_update_model.vendor_number,
            asset_update_model.bus_type,
            asset_update_model.name,
            asset_update_model.asset_type,
            asset_update_model.firmware_version,
            asset_update_model.hardware_version,
            asset_update_model.visa_resource_name,
            asset_update_model.temperature_sensors,
            asset_update_model.supports_self_calibration,
            asset_update_model.supports_external_calibration,
            asset_update_model.self_calibration,
            asset_update_model.is_ni_asset,
            asset_update_model.location,
            asset_update_model.external_calibration
        )
        asset_create_models.append(asset_create_model)
    return asset_update_models
